'use strict';

/**
 * Integration test for aws-stream-consumer-core/persisting.js to test loading of tracked state for a batch sourced from a
 * Kinesis stream.
 * @author Byron du Preez
 */

const test = require('tape');

// The test subject
const persisting = require('aws-stream-consumer-core/persisting');
const loadBatchStateFromDynamoDB = persisting.loadBatchStateFromDynamoDB;

const kinesisProcessing = require('../../kinesis-processing');

const Batch = require('aws-stream-consumer-core/batch');

const Settings = require('aws-stream-consumer-core/settings');
const StreamType = Settings.StreamType;

const sequencing = require('aws-stream-consumer-core/sequencing');

const streamEvents = require('aws-core-utils/stream-events');

const taskUtils = require('task-utils');
const TaskDef = require('task-utils/task-defs');
const Task = require('task-utils/tasks');
// const taskStates = require('task-utils/task-states');
// const TaskState = taskStates.TaskState;
// const core = require('task-utils/core');
// const StateType = core.StateType;
// const cleanCompletedState = JSON.parse(JSON.stringify(taskStates.instances.Completed));

const logging = require('logging-utils');
const LogLevel = logging.LogLevel;

const Promises = require('core-functions/promises');

const strings = require('core-functions/strings');
const stringify = strings.stringify;
const isBlank = strings.isBlank;

const tries = require('core-functions/tries');
//const Try = tries.Try;
const Success = tries.Success;
// const Failure = tries.Failure;

//Gets the AWS region to use from an AWS_REGION environment variable
process.env.AWS_REGION = 'eu-west-1';
const awsRegion = process.env.AWS_REGION;
if (isBlank(awsRegion)) {
  throw new Error('Set the AWS region you want to use as an AWS_REGION environment variable')
}

//Gets the AWS account ID to use from an AWS_ACCOUNT_ID environment variable
process.env.AWS_ACCOUNT_ID = 'XXXXXXXXXX';
const accountNumber = process.env.AWS_ACCOUNT_ID;
if (isBlank(accountNumber)) {
  throw new Error('Set the AWS Account ID you want to use as an AWS_ACCOUNT_ID environment variable')
}

const samples = require('../samples');

const eventSourceARN = `arn:aws:kinesis:${awsRegion}:${accountNumber}:stream/MyStream_DEV`;

function createContext(sequencingRequired, idPropertyNames, keyPropertyNames, seqNoPropertyNames) {
  idPropertyNames = idPropertyNames ? idPropertyNames : ['id1', 'id2'];
  keyPropertyNames = sequencingRequired ? keyPropertyNames ? keyPropertyNames : ['k1', 'k2'] : [];
  seqNoPropertyNames = seqNoPropertyNames ? seqNoPropertyNames : ['n1', 'n2', 'n3'];
  const context = {
    region: awsRegion,
    stage: 'dev'
  };
  const streamProcessingOptions = {
    streamType: StreamType.kinesis,
    sequencingRequired: sequencingRequired,
    sequencingPerKey: sequencingRequired && (keyPropertyNames && keyPropertyNames.length > 0),
    consumerIdSuffix: undefined,
    maxNumberOfAttempts: 2,
    idPropertyNames: idPropertyNames,
    keyPropertyNames: keyPropertyNames,
    seqNoPropertyNames: seqNoPropertyNames,
    batchStateTableName: 'TEST_StreamConsumerBatchState'
  };
  logging.configureLogging(context, {logLevel: LogLevel.TRACE});

  kinesisProcessing.configureDefaultKinesisStreamProcessing(context, streamProcessingOptions, undefined,
    require('../../default-kinesis-options.json'), undefined, undefined, true);

  context.streamProcessing.consumerId = `my-function:${context.stage}`;

  // Set the task factory to use on the context (after configuring logging, so that the context can be used as a logger too)
  taskUtils.configureTaskFactory(context, {logger: context}); //, {returnMode: taskUtils.ReturnMode.NORMAL});
  return context;
}

function processOne(message, batch, context) {
  context.trace(`Executing processOne on message (${batch.states.get(message).msgDesc})`);
  return Promise.resolve(batch.states.get(message).eventID);
}

function processAll(batch, context) {
  const messages = batch.messages;
  context.trace(`Executing processAll on batch (${batch.shardOrEventID})`);
  return Promise.resolve(messages.map(m => batch.states.get(m).eventID));
}

function discardUnusableRecord(unusableRecord, batch, context) {
  context.trace(`Simulating execution of discardUnusableRecord on record (${batch.states.get(unusableRecord).recDesc})`);
  return Promise.resolve(unusableRecord);
}

function discardRejectedMessage(rejectedMessage, batch, context) {
  context.trace(`Simulating execution of discardRejectedMessage on message (${batch.states.get(rejectedMessage).msgDesc})`);
  return Promise.resolve(rejectedMessage);
}

function kinesisFixture(sequencingRequired) {
  const context = createContext(sequencingRequired, ['id1', 'id2'], ['k1', 'k2'], ['n1', 'n2', 'n3', 'n4']);

  const processOneTaskDef = TaskDef.defineTask(processOne.name, processOne);
  const processAllTaskDef = TaskDef.defineTask(processAll.name, processAll);

  context.streamProcessing.discardUnusableRecord = discardUnusableRecord;
  context.streamProcessing.discardRejectedMessage = discardRejectedMessage;

  const shardId = sequencingRequired ? 'shardId-777777777777' : 'shardId-888888888888';

  // Message 1 - usable
  const shardId1 = shardId;
  const seqNo1 = '49545115243490985018280067714973144582180062593244200961';
  const [msg1, record1] = samples.sampleKinesisMessageAndRecord(shardId1, seqNo1, eventSourceARN, 'ID1', '1001', 'ABC', 10, 1, 100, '10000000000000000000001', '2017-01-17T23:59:59.001Z');
  record1.kinesis.partitionKey = 'abfac769-e349-4d16-90d0-b857a578971f';

  // Message 2 - "usable", but with rejected task
  const shardId2 = shardId;
  const seqNo2 = '49545115243490985018280067714973144582180062593244200962';
  const [msg2, record2] = samples.sampleKinesisMessageAndRecord(shardId2, seqNo2, eventSourceARN, 'ID1', '1002', 'ABC', 10, 1, 100, '10000000000000000000001', '2017-01-17T23:59:59.002Z');
  record2.kinesis.partitionKey = 'b660092a-27d4-46b6-aa83-cade97ed6099';

  // Message 3 - "unusable"
  const shardId3 = shardId;
  const seqNo3 = '49545115243490985018280067714973144582180062593244200963';
  // noinspection JSUnusedLocalSymbols
  const [msg3, record3] = samples.sampleKinesisMessageAndRecord(shardId3, seqNo3, eventSourceARN, 'ID1', '1003', 'ABC', 10, 1, 100, '10000000000000000000003', '2017-01-17T23:59:59.003Z');
  record3.kinesis.partitionKey = '8e2893e9-0134-45d1-8f82-1c1f8c732cf3';

  // Message 4 - also usable
  const shardId4 = shardId;
  const seqNo4 = '49545115243490985018280067714973144582180062593244200964';
  const [msg4, record4] = samples.sampleKinesisMessageAndRecord(shardId4, seqNo4, eventSourceARN, 'ID1', '1004', 'ABC', 10, 1, 100, '10000000000000000000001', '2017-01-17T23:59:59.004Z');
  record4.kinesis.partitionKey = 'c291e2aa-87b0-48bf-85ff-500bb1f6041e';

  // Message 5 - also "usable", but with rejected task
  const shardId5 = shardId;
  const seqNo5 = '49545115243490985018280067714973144582180062593244200965';
  const [msg5, record5] = samples.sampleKinesisMessageAndRecord(shardId5, seqNo5, eventSourceARN, 'ID1', '1005', 'ABC', 10, 1, 100, '10000000000000000000001', '2017-01-17T23:59:59.005Z');
  record5.kinesis.partitionKey = 'da12eda8-ed73-4dbf-b54b-73dc1bd16a0a';

  // Unusable record 1
  const badRec1 = {badApple: true};

  // Unusable record 2
  // noinspection UnnecessaryLocalVariableJS
  const badRec2 = record3;

  return {
    context: context,
    cancellable: {},
    processOneTaskDef: processOneTaskDef,
    processAllTaskDef: processAllTaskDef,
    shardId: shardId,

    shardId1: shardId1,
    eventID1: record1.eventID,
    seqNo1: seqNo1,
    msg1: msg1,
    record1: record1,

    shardId2: shardId2,
    eventID2: record2.eventID,
    seqNo2: seqNo2,
    msg2: msg2,
    record2: record2,

    badRec1: badRec1,
    badRec2: badRec2,

    shardId3: shardId3,
    eventID3: record3.eventID,
    seqNo3: seqNo3,
    msg3: msg3,
    record3: record3,

    shardId4: shardId4,
    eventID4: record4.eventID,
    seqNo4: seqNo4,
    msg4: msg4,
    record4: record4,

    shardId5: shardId5,
    eventID5: record5.eventID,
    seqNo5: seqNo5,
    msg5: msg5,
    record5: record5
  };
}

function executeTasks(f, batch, t) {
  // Message 1 task 1 (process one task)
  const msg1Task1 = batch.getProcessOneTask(f.msg1, processOne.name);
  const m1p1 = msg1Task1 ? msg1Task1.execute(f.msg1, batch, f.context) : Promise.resolve(undefined);

  if (batch.messages.indexOf(f.msg1) !== -1 && batch.taskDefs.processOneTaskDefs.length > 0) {
    t.ok(msg1Task1, `msg1Task1 must exist`);
  }

  // Message 2 task 1 (process one task) - execute & then mark as rejected
  const msg2Task1 = batch.getProcessOneTask(f.msg2, processOne.name);
  const m2p1 = msg2Task1 ? msg2Task1.execute(f.msg2, batch, f.context) : Promise.resolve(undefined);

  if (batch.messages.indexOf(f.msg2) !== -1 && batch.taskDefs.processOneTaskDefs.length > 0) {
    t.ok(msg2Task1, `msg2Task1 must exist`);
  }

  const m2p1b = m2p1.then(res => {
    if (msg2Task1) {
      msg2Task1.reject('We\'re NOT gonna take it!', new Error('Twisted'));
    }
    return res;
  });

  // Message 4 task 1 (process one task)
  const msg4Task1 = batch.getProcessOneTask(f.msg4, processOne.name);
  const m4p1 = msg4Task1 ? msg4Task1.execute(f.msg4, batch, f.context) : Promise.resolve(undefined);

  if (batch.messages.indexOf(f.msg4) !== -1 && batch.taskDefs.processOneTaskDefs.length > 0) {
    t.ok(msg4Task1, `msg4Task1 must exist`);
  }

  // Message 5 task 1 (process one task) - execute & then mark as rejected
  const msg5Task1 = batch.getProcessOneTask(f.msg5, processOne.name);
  const m5p1 = msg5Task1 ? msg5Task1.execute(f.msg5, batch, f.context) : Promise.resolve(undefined);

  if (batch.messages.indexOf(f.msg5) !== -1 && batch.taskDefs.processOneTaskDefs.length > 0) {
    t.ok(msg5Task1, `msg5Task1 must exist`);
  }

  const m5p1b = m5p1.then(res => {
    if (msg5Task1) {
      msg5Task1.reject('No, we ain\'t gonna take it!', new Error('Twisted too'));
    }
    return res;
  });

  // Batch task 1 (process all task)
  const batchTask1 = batch.getProcessAllTask(batch, processAll.name);
  const b1p1 = batchTask1 ? batchTask1.execute(batch, f.context) : Promise.resolve(undefined);

  if (batch.messages.length > 0 && batch.taskDefs.processAllTaskDefs.length > 0) {
    t.ok(batchTask1, `batchTask1 must exist`);

    t.ok(batchTask1.isMasterTask(), `batchTask1 must be a master task`);

    // Message 1 task 2 (process all task)
    if (batch.messages.indexOf(f.msg1) !== -1) {
      const msg1Task2 = batch.getProcessAllTask(f.msg1, processAll.name);
      t.ok(msg1Task2, `msg1Task2 must exist`);
    }

    // Message 2 task 2 (process all task)
    if (batch.messages.indexOf(f.msg2) !== -1) {
      const msg2Task2 = batch.getProcessAllTask(f.msg2, processAll.name);
      t.ok(msg2Task2, `msg2Task2 must exist`);
    }

    // Message 4 task 2 (process all task)
    if (batch.messages.indexOf(f.msg4) !== -1) {
      const msg4Task2 = batch.getProcessAllTask(f.msg4, processAll.name);
      t.ok(msg4Task2, `msg4Task2 must exist`);
    }

    // Message 5 task 2 (process all task)
    if (batch.messages.indexOf(f.msg5) !== -1) {
      const msg5Task2 = batch.getProcessAllTask(f.msg5, processAll.name);
      t.ok(msg5Task2, `msg5Task2 must exist`);
    }
  }

  const d1 = batch.discardUnusableRecords(f.cancellable, f.context);

  const d2 = Promise.all([m2p1b, m5p1b, b1p1]).then(() => batch.discardRejectedMessages(f.cancellable, f.context));

  return Promises.every([m1p1, m2p1, m4p1, m5p1, m2p1b, m5p1b, b1p1, d1, d2], f.cancellable, f.context);
}

// =====================================================================================================================
// loadBatchStateFromDynamoDB - kinesis - sequencingRequired true
// =====================================================================================================================

test('loadBatchStateFromDynamoDB for kinesis - sequencingRequired true - 2 messages & 2 unusable records', t => {
  const f = kinesisFixture(true);
  const taskFactory = f.context.taskFactory;

  // f.context.dynamoDBDocClient = ...

  const task = taskFactory.createTask(TaskDef.defineTask(loadBatchStateFromDynamoDB.name, loadBatchStateFromDynamoDB));

  const batch = new Batch([f.badRec1, f.badRec2, f.record1, f.record2], [f.processOneTaskDef], [f.processAllTaskDef], f.context);
  f.uRec1 = batch.addUnusableRecord(f.badRec1, undefined, 'Bad apple', f.context);
  f.uRec2 = batch.addUnusableRecord(f.badRec2, undefined, 'Unparseable', f.context);
  batch.addMessage(f.msg1, f.record1, undefined, f.context);
  batch.addMessage(f.msg2, f.record2, undefined, f.context);

  const expectedShardId = 'shardId-777777777777';
  batch.messages.forEach((msg, i) => {
    const state = batch.states.get(msg);
    const shardId = state.eventID ? streamEvents.getKinesisShardIdFromEventID(state.eventID) : undefined;
    t.equal(shardId, expectedShardId, `batch message[${i}] shardId must be ${expectedShardId}`)
  });

  batch.reviveTasks(f.context);

  // Simulate execution of tasks
  const promise = Promise.resolve(undefined); //executeTasks(f, batch, t);

  promise.then(
    () => {
      const executeResult = task.execute(batch, f.context);
      return executeResult.then(
        result => {
          t.pass(`loadBatchStateFromDynamoDB must succeed - result ${stringify(result && result.Item)}`);
          t.ok(task.started, `Task (${task.name}) must be started`);
          t.equal(task.result, undefined, `Task (${task.name}) result must be undefined`);
          t.equal(task.error, undefined, `Task (${task.name}) error must be undefined`);

          t.deepEqual(task.outcome, new Success(executeResult), `Task (${task.name}) outcome must be Success(executeResult)`);
          t.equal(task.outcome.value, executeResult, `Task (${task.name}) outcome.value must be executeResult`);

          t.ok(task.donePromise, `Task (${task.name}) donePromise must not be undefined`);
          t.ok(task.donePromise instanceof Promise, `Task (${task.name}) donePromise must be a Promise`);
          return task.donePromise.then(
            doneResult => {
              t.ok(doneResult, `Task (${task.name}) donePromise result must be defined`);
              t.equal(doneResult, result, `Task (${task.name}) donePromise result must be executeResult result`);

              t.ok(task.completed, `Task (${task.name}) must still be completed`);
              t.equal(task.result, result, `Task (${task.name}) result must still be result`);
              t.end();
            },
            err => {
              t.fail(`Task (${task.name}) donePromise must NOT reject`, err);
              t.end();
            }
          );
        },
        err => {
          t.fail(`loadBatchStateFromDynamoDB must NOT fail`, err);
          t.end();
        }
      );
    },
    err => {
      t.end(err);
    }
  );
});

// // =====================================================================================================================
// // loadBatchStateFromDynamoDB - kinesis - sequencingRequired false
// // =====================================================================================================================
//
// test('loadBatchStateFromDynamoDB for kinesis - sequencingRequired false - 2 messages & 2 unusable records', t => {
//   const f = kinesisFixture(false);
//   const taskFactory = f.context.taskFactory;
//
//   // f.context.dynamoDBDocClient = ...
//
//   const task = taskFactory.createTask(TaskDef.defineTask(loadBatchStateFromDynamoDB.name, loadBatchStateFromDynamoDB));
//
//   const batch = new Batch([f.badRec1, f.badRec2, f.record1, f.record2], [f.processOneTaskDef], [f.processAllTaskDef], f.context);
//   f.uRec1 = batch.addUnusableRecord(f.badRec1, undefined, 'Bad apple', f.context);
//   f.uRec2 = batch.addUnusableRecord(f.badRec2, undefined, 'Unparseable', f.context);
//   batch.addMessage(f.msg1, f.record1, undefined, f.context);
//   batch.addMessage(f.msg2, f.record2, undefined, f.context);
//
//   const expectedShardId = 'shardId-888888888888';
//   batch.messages.forEach((msg, i) => {
//     const state = batch.states.get(msg);
//     const shardId = state.eventID ? streamEvents.getKinesisShardIdFromEventID(state.eventID) : undefined;
//     t.equal(shardId, expectedShardId, `batch message[${i}] shardId must be ${expectedShardId}`)
//   });
//
//   batch.reviveTasks(f.context);
//
//   // Simulate execution of tasks
//   const promise = Promise.resolve(undefined); // executeTasks(f, batch, t);
//
//   promise.then(
//     () => {
//       const executeResult = task.execute(batch, f.context);
//       return executeResult.then(
//         result => {
//           t.pass(`loadBatchStateFromDynamoDB must succeed - result ${stringify(result && result.Item)}`);
//           t.ok(task.started, `Task (${task.name}) must be started`);
//           t.equal(task.result, undefined, `Task (${task.name}) result must be undefined`);
//           t.equal(task.error, undefined, `Task (${task.name}) error must be undefined`);
//
//           t.deepEqual(task.outcome, new Success(executeResult), `Task (${task.name}) outcome must be Success(executeResult)`);
//           t.equal(task.outcome.value, executeResult, `Task (${task.name}) outcome.value must be executeResult`);
//
//           t.ok(task.donePromise, `Task (${task.name}) donePromise must not be undefined`);
//           t.ok(task.donePromise instanceof Promise, `Task (${task.name}) donePromise must be a Promise`);
//           return task.donePromise.then(
//             doneResult => {
//               t.ok(doneResult, `Task (${task.name}) donePromise result must be defined`);
//               // t.equal(doneResult, data, `Task (${task.name}) donePromise result must be data`);
//               t.equal(doneResult, result, `Task (${task.name}) donePromise result must be executeResult result`);
//
//               t.ok(task.completed, `Task (${task.name}) must still be completed`);
//               t.equal(task.result, result, `Task (${task.name}) result must still be result`);
//               t.end();
//             },
//             err => {
//               t.fail(`Task (${task.name}) donePromise must NOT reject`, err);
//               t.end();
//             }
//           );
//         },
//         err => {
//           t.fail(`loadBatchStateFromDynamoDB must NOT fail`, err);
//           t.end();
//         }
//       );
//     },
//     err => {
//       t.fail(`promise must NOT fail`, err);
//       t.end(err);
//     }
//   );
// });